import logging

from django.core.management.base import BaseCommand
from django.test import override_settings

from dsbot.client import BotClient
from dsbot.conf import settings

try:
    from importlib_metadata import entry_points
except ImportError:
    from importlib.metadata import entry_points


class Command(BaseCommand):
    def add_arguments(self, parser):
        slack = parser.add_argument_group("Slack arguments")
        slack.add_argument("--token", default=settings.SLACK_TOKEN, help="Slack token")
        slack.add_argument("--timeout", default=30, type=int)
        slack.add_argument("--ping-interval", default=30, type=int)

        celery = parser.add_argument_group("Celery Arguments")
        celery.add_argument("--eager", action="store_true")

    def handle(self, verbosity, eager, **options):
        logging.root.setLevel(
            {
                0: logging.ERROR,
                1: logging.WARNING,
                2: logging.INFO,
                3: logging.DEBUG,
            }.get(verbosity)
        )

        ch = logging.StreamHandler()
        formatter = logging.Formatter(logging.BASIC_FORMAT)
        ch.setFormatter(formatter)

        logging.root.addHandler(ch)

        for entry in entry_points(group="dsbot.commands"):
            try:
                entry.load()
            except ImportError:
                logging.exception("Error loading %s", entry)
            else:
                logging.info("Loaded %s", entry)

        with override_settings(CELERY_TASK_ALWAYS_EAGER=eager):
            BotClient(
                token=options["token"],
                timeout=options["timeout"],
                ping_interval=options["ping_interval"],
            ).start()



from unittest import mock

from slack_sdk.web.slack_response import SlackResponse

from django.test import TestCase

from dsbot import exceptions, tasks


def mock_error(message="mock_error", **kwargs):
    kwargs.setdefault("ok", False)
    return exceptions.SlackApiError(
        message,
        SlackResponse(
            client=None,
            http_verb="POST",
            api_url="example.com",
            req_args={},
            data=kwargs,
            headers={},
            status_code=400,
        ),
    )


class TestExceptions(TestCase):
    @mock.patch("slack_sdk.web.client.WebClient.api_call")
    def test_not_in_channel(self, client: mock.MagicMock):
        with self.assertRaises(exceptions.NotInChannel):
            client.side_effect = mock_error(error="not_in_channel")
            tasks.api_call("chat.postMessage", channel="test_channel")

        with self.assertRaises(exceptions.ArchiveException):
            client.side_effect = mock_error(error="is_archived")
            tasks.api_call("chat.postMessage", channel="test_channel")

        with self.assertRaises(exceptions.ChannelNotFound):
            client.side_effect = mock_error(error="channel_not_found")
            tasks.api_call("chat.postMessage", channel="test_channel")



import argparse
import os
import sys

import django
from django.conf import settings
from django.test.utils import get_runner

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.example.settings"

parser = argparse.ArgumentParser()
# Arguments from django-admin test
parser.add_argument(
    "-v",
    "--verbosity",
    default=1,
    type=int,
    choices=[0, 1, 2, 3],
    help="Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output",
)
parser.add_argument(
    "args",
    metavar="test_label",
    nargs="*",
    help="Module paths to test; can be modulename, modulename.TestCase or modulename.TestCase.test_method",
)
options = parser.parse_args()


django.setup()
TestRunner = get_runner(settings)
test_runner = TestRunner(verbosity=options.verbosity)
failures = test_runner.run_tests(options.args)
sys.exit(bool(failures))



from slack_sdk.errors import SlackApiError


class BaseException(Exception):
    pass


class CommandError(BaseException):
    pass


class ArgumentError(CommandError):
    pass


# Custom slack exceptions


class ChannelError(SlackApiError):
    def __init__(self, response, **kwargs):
        self.channel = kwargs.get("channel", "Unknown")
        super().__init__(
            f"{response.data['error']} for {self.channel}",
            response,
        )


class NotInChannel(ChannelError):
    pass


class ArchiveException(ChannelError):
    pass


class ChannelNotFound(ChannelError):
    pass


# https://api.slack.com/methods/chat.postMessage#errors
channel_errors = {
    "channel_not_found": ChannelNotFound,
    "not_in_channel": NotInChannel,
    "is_archived": ArchiveException,
}


def cast_slack_exception(exception: SlackApiError, **kwargs):
    if exception.response.data["error"] in channel_errors:
        raise channel_errors[exception.response.data["error"]](
            response=exception.response, **kwargs
        ) from exception
    raise exception


class SlackException(BaseException):
    def __init__(self, response):
        self.data = response

    def __str__(self):
        return self.data["error"]

    def __repr__(self):
        return "SlackException(%s)" % self



import logging
import re
from html import unescape

MENTION_REGEX = re.compile("^<@(|[WU].+?)>(.*)")
LINK_REGEX = re.compile(r"<(http.*?)(\|.*?)?>", re.DOTALL)

logger = logging.getLogger(__name__)


def parse_direct_mention(message_text):
    """
    Finds a direct mention (a mention that is at the beginning) in message text
    and returns the user ID which was mentioned. If there is no direct mention, returns None
    """
    matches = MENTION_REGEX.search(message_text)
    # the first group contains the username, the second group contains the remaining message
    return (matches.group(1), matches.group(2).strip()) if matches else (None, None)


# Parse URLs
# https://api.slack.com/docs/message-formatting#linking_to_urls
def parse_links(message_text):
    for m in LINK_REGEX.findall(message_text):
        logger.debug("Found match %s", m)
        # Links in text sometimes have their enti
        yield unescape(m[0])



from celery import shared_task
from slack_sdk.web.client import WebClient

from dsbot import exceptions
from dsbot.conf import settings

client = WebClient(token=settings.SLACK_TOKEN)


# Wrapped version of Slack API Calll
# We want to make it easy to rate limit our calls to slack by wrapping
# it as a shared_task.
@shared_task(rate_limit=settings.SLACK_RATE_LIMIT)
def api_call(*args, **kwargs):
    try:
        return client.api_call(*args, json=kwargs).data
    except exceptions.SlackApiError as e:
        exceptions.cast_slack_exception(e, **kwargs)



"""
These are the core commands and events that typically require a greater
knowledge about the internals of the bot itself
"""

import logging

from django.template.loader import render_to_string

from dsbot.client import BotClient
from dsbot.util import parse_direct_mention

logger = logging.getLogger(__name__)


@BotClient.cmd("^help")
async def cmd_help(rtm_client, web_client, data, **kwargs):
    """
    help - Show list of commands

    The help command loops through all registered dispatch commands and
    formats their help output via django's render_to_string method
    """
    return web_client.chat_postEphemeral(
        channel=data["channel"],
        as_user=True,
        text=render_to_string(
            "slack/response/help.txt", {"mapping": rtm_client._commands}
        ).strip(),
        user=data["user"],
    )


@BotClient.run_on(event="open")
async def get_team_data(rtm_client, data, **payload):
    rtm_client.user_id = data["self"]["id"]


@BotClient.run_on(event="message")
async def command_checker(web_client, rtm_client, data, **kwargs):
    """
    Basic command dispatch

    We want our bot to be a good citizen, so in public channels
    we only want it to respond to a direct @-mention

    In the case of direct private messages, we don't want to
    require @-mention, since it's obvious it's a command directly
    to the bot
    """
    if "subtype" in data:
        # https://api.slack.com/events/message#message_subtypes
        logger.debug("command_checker skips subtypes %s", data["subtype"])
        return

    # Check to see if this is an @ message
    user_id, message = parse_direct_mention(data["text"])

    if user_id == rtm_client.user_id:
        return await rtm_client._dispatch_command(message, data)
    # Or a direct message
    if data["channel"].startswith("D"):
        return await rtm_client._dispatch_command(data["text"], data)

    # Otherwise we ignore it to be a good citizen
    logger.debug("Ignoring non @ message in public channel %s", data)



from django.test import TestCase

from dsbot import util


class TestUtilPackage(TestCase):
    def test_parse(self):
        result = util.parse_links(
            """
        Test Text
        <https://example.com?one=1&amp;two=2>\r\n<https://another.example.com>
        Again
        """
        )
        self.assertEqual(
            list(result),
            ["https://example.com?one=1&two=2", "https://another.example.com"],
        )

    def test_direct_message(self):
        not_a_mention = "This is not a mention"
        result = util.parse_direct_mention(not_a_mention)
        self.assertEqual(result, (None, None))

        cc_mention = "CC <@Usomehting>"
        result = util.parse_direct_mention(cc_mention)
        self.assertEqual(result, (None, None))

        direct_mention = "<@Usomehting> Check this out"
        result = util.parse_direct_mention(direct_mention)
        self.assertEqual(result, ("Usomehting", "Check this out"))



"""
Base Bot Class

The base bot class is mostly concerned with maintaining
the connection to Slack, and then dispatching events to
the Dispatcher

A few convenencie functions used by commands are also
added to the bot class
"""


import concurrent
import inspect
import logging
import re

import slack_sdk.rtm
import slack_sdk.web

from .exceptions import CommandError

logger = logging.getLogger(__name__)


class BotClient(slack_sdk.rtm.RTMClient):
    user_id = None
    _commands = []

    @classmethod
    def cmd(cls, key):
        """A decorator to store and link a callback to an event."""

        def decorator(callback):
            logger.debug("Registering %s %s", key, callback)
            cls._validate_callback(callback)
            cls._commands.append(
                {
                    "key": key,
                    "re": re.compile(key),
                    "func": callback,
                    "help": callback.__doc__.strip().split("\n")[0],
                }
            )
            return callback

        return decorator

    async def _dispatch_command(self, command, data=None):
        for cmd in self._commands:
            match = cmd["re"].match(command)
            if match:
                try:
                    if inspect.iscoroutinefunction(cmd["func"]):
                        logger.debug("Running %(key)s %(func)s as async", cmd)
                        return await cmd["func"](
                            rtm_client=self,
                            web_client=self._web_client,
                            data=data,
                            match=match,
                        )
                    else:
                        logger.debug("Running %(key)s %(func)s as thread", cmd)
                        self._cmd_in_thread(cmd["func"], data=data, match=match)
                except CommandError as e:
                    logger.warning("Command Error")
                    return self._web_client.chat_postEphemeral(
                        as_user=True,
                        channel=data["channel"],
                        user=data["user"],
                        attachments=[
                            {
                                "color": "warning",
                                "title": "Command Error",
                                "text": str(e),
                            }
                        ],
                    )
                except Exception as e:
                    logger.exception("Unknown Error")
                    return self._web_client.chat_postEphemeral(
                        as_user=True,
                        channel=data["channel"],
                        user=data["user"],
                        attachments=[
                            {
                                "color": "danger",
                                "title": "Unknown Error",
                                "text": str(e),
                            }
                        ],
                    )

    def _cmd_in_thread(self, callback, **kwargs):
        """Execute the callback in another thread. Wait for and return the results."""
        web_client = slack_sdk.web.WebClient(
            token=self.token,
            base_url=self.base_url,
            ssl=self.ssl,
            proxy=self.proxy,
            headers=self.headers,
        )
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(
                callback, rtm_client=self, web_client=web_client, **kwargs
            )

            while future.running():
                pass

            future.result()



import logging
from functools import wraps

from dsbot.conf import settings

logger = logging.getLogger(__name__)


def ignore_users(ignore_user_list=settings.SLACK_IGNORE_USERS):
    def _outer(func):
        @wraps(func)
        def _inner(*args, data, **kwargs):
            if data.get("user") in ignore_user_list:
                logger.debug("Ignoring user %s", data["user"])
            else:
                return func(*args, data=data, **kwargs)

        return _inner

    return _outer


def ignore_bots(func):
    @wraps(func)
    def _inner(*args, data, **kwargs):
        subtype = data.get("subtype", "")
        if subtype not in ["bot_message", ""]:
            # https://api.slack.com/events/message/bot_message
            # We only care about bot_message or message without a subtype
            logger.debug("Skipping bot message %s", subtype)
        else:
            return func(*args, data=data, **kwargs)

    return _inner



from django.conf import settings as django_settings

DEFAULTS = {
    "SLACK_IGNORE_USERS": [],
    # Since we typically deploy 2 workers and we don't want to go too
    # much over the 1 message / second rate limit, we set our rate limit
    # to 0.5 to be equivilant to 1 message / 2 seconds
    # https://api.slack.com/docs/rate-limits#tier_t5
    "SLACK_RATE_LIMIT": 0.5,
}


class SettingsWrapper:
    def __getattr__(self, key):
        if key in DEFAULTS:
            return getattr(django_settings, key, DEFAULTS[key])
        return getattr(django_settings, key)


settings = SettingsWrapper()
